/*
  Color.h - Library for controlling RGB objects
  Created by Jesse R. Castro 2013-02-03
  Reworked by Maxime Fouillat in 2022-06
    Switch to RGB-based instead of HSV-based
    Added two-way redundancy to and from RGB, fixes empty RGB values
    Converted properties to interger and byte logic instead of floating point for lesser memory usage
    Fixed operator overloading for RGB compatibility
    Reorganized public and private parts of the class
*/

#include "Color.h"


//Constructors 
Color::Color(int pRed, int pGreen, int pBlue){
	setRGB(pRed, pGreen, pBlue);
  rgb2hsv();
}

/*
Color::Color(int pHue, int pSaturation, int pValue){
	setHSV(pHue, pSaturation, pValue);
  hsv2rgb();
}*/

Color::Color(const Color& pColor){
	*this = pColor;
}
Color::Color(){
	setHSV(0,0,0);
  setRGB(0,0,0);
}
//setters


void Color::setHue(int pHue){
	//Note, of these hue cycles, so we have to make sure that works
	//abs gets rid of any negatives
	//the val - floor(val) will take return only decimal portions
	hue = abs(pHue) - floor(abs(pHue));
	
}
void Color::setSaturation(int pSaturation){
	if(pSaturation < 0){
		saturation = 0;
	}else if(pSaturation > 255){
		saturation = 255;
	}else{
		saturation = pSaturation;
	}
}
void Color::setValue(int pValue){
	if(pValue < 0){
		value = 0;
	}else if(pValue > 255){
		value = 255;
	}else{
		value = pValue;
	}
}

void Color::setHSV(int pHue, int pSaturation, int pValue){
	setHue(pHue);
	setSaturation(pSaturation);
	setValue(pValue);
  hsv2rgb();
}

void Color::setRed(int pRed){
	if(pRed < 0){
		red = 0;
	}else if(pRed > 255){
		red = 255;
	}else{
		red = pRed;
	}
}
void Color::setBlue(int pBlue){
	if(pBlue < 0){
		blue = 0;
	}else if(pBlue > 255){
		blue = 255;
	}else{
		blue = pBlue;
	}
}
void Color::setGreen(int pGreen){
	if(pGreen < 0){
		green = 0;
	}else if(pGreen > 255){
		green = 255;
	}else{
		green = pGreen;
	}
}
void Color::setRGB(int pRed, int pGreen, int pBlue){
	setRed(pRed);
	setGreen(pGreen);
	setBlue(pBlue);
  rgb2hsv();
}

//getters
int Color::getHue() const{
	return hue;
}
int Color::getSaturation() const{
	return saturation;
}
int Color::getValue() const{
	return value;
}
int Color::getRed(){
	return red;
}
int Color::getGreen(){
	return green;
}
int Color::getBlue(){
	return blue;
}

//action methods
void Color::turnOff(){
	setValue(0);
}


void Color::rgb2hsv(){
	float r = (red + 0)/255.0;
	float g = (green + 0)/255.0; 
	float b = (blue + 0)/255.0;
			
	float max = max(max(r, g), b);
	float min = min(min(r, g), b);
	
	float h = max;
	float s = max;
	float v = max;

	float d = max - min;
	s = max == 0 ? 0 : d / max;

	if(max == min){
		h = 0; // achromatic
	}else{
		if(max == r){
			h = (g - b) / d + (g < b ? 6 : 0);
		}else if (max == g){
			h = (b - r) / d + 2; 
		}else{
			h = (r - g) / d + 4; 
		}
		h /= 6;
	}

	setHue(int(h * 255));
	setSaturation(int(s * 255));
	setValue(int(v * 255));
}
void Color::hsv2rgb(){
	float r, g, b;

	float h = hue / 255;        //convertion to [0,1] range from [0,255]
	float s = saturation / 255;
	float v = value / 255;
	
	int i = floor(h * 6);	
	float f = h * 6 - i;
	float p = v * (1 - s);
	float q = v * (1 - f * s);
	float t = v * (1 - (1 - f) * s);

	switch(i % 6){
		case 0: r = v, g = t, b = p; break;
		case 1: r = q, g = v, b = p; break;
		case 2: r = p, g = v, b = t; break;
		case 3: r = p, g = q, b = v; break;
		case 4: r = t, g = p, b = v; break;
		case 5: r = v, g = p, b = q; break;
	}

	setRed(round(r*255));
	setGreen(round(g*255));
	setBlue(round(b*255));
}

float Color::getMin(){
	float r = (red + 0)/255.0;
	float g = (green + 0)/255.0;
	float b = (blue + 0)/255.0;
	
	if(r < g){
		if(r < b){
			return r;
		}else{
			return b;
		}
	}else{
		if(g < b){
			return g;
		}else{
			return b;
		}
	}
}
float Color::getMax(){
	float r = (red + 0)/255.0;
	float g = (green + 0)/255.0;
	float b = (blue + 0)/255.0;
	
	if(r > g){
		if(r > b){
			return r;
		}else{
			return b;
		}
	}else{
		if(g > b){
			return g;
		}else{
			return b;
		}
	}
}
int Color::floor(float target){
	int returnVal = round(target);
	if(returnVal > target){
		returnVal--;
	}
	return returnVal;
}



//operator overloads
Color& Color::operator=(const Color &pColor){
	setRGB(pColor.red, pColor.green, pColor.blue);	
	return *this;
}

/** 
 * Comparison == operator overload.
 * @author Jesse R. Castro
 * @param pFoo An instance of the same type as this object.
 * @return A boolean value describing whether the objects are equivalent or not.
 */
bool Color::operator==(const Color &pColor) const {
	if(
		(hue == pColor.hue) &&
		(saturation == pColor.saturation) &&
		(value == pColor.value)
	){
		return true;
	}else{
		return false;
	}
}
/** 
 * Comparison != operator overload.
 * @author Jesse R. Castro
 * @param pFoo An instance of the same type as this object.
 * @return A boolean value describing the inverse of whether the objects are equivalent or not.
 */
bool Color::operator!=(const Color &pColor) const {
	return !(*this == pColor);
}